#!/bin/bash
#
# This script checks unisolvence of a symmetric point-set for a hierarchical
# polynomial space by symbolic evaluation of det(f_i(p_j)) using SageMath or
# Mathematica.
# 
# Note: for H1-conforming point-sets, because of the block triangular structure
#	of the characteristic matrix f_i(p_j), we only need to check the
#	unisolvence for the subspace of interior functions.
#
# $Id: quadrule-ml-symbolic-check.sh,v 1.37 2016/10/14 06:49:17 zlb Exp $

# PRINT="true": print SageMath or Mathematica scripts to /dev/stderr
PRINT_DEFAULT="false"
test "$PRINT" != "true" -a "$PRINT" != "false" && PRINT=$PRINT_DEFAULT

# UNSYM="true": unsymmetric, UNSYM="false": symmetric
UNSYM_DEFAULT="false"
test "$UNSYM" != "true" -a "$UNSYM" != "false" && UNSYM=$UNSYM_DEFAULT

# NUMERIC="true": numeric evaluation, NUMERIC="false": symbolic evaluation
NUMERIC_DEFAULT="false"
test "$NUMERIC" != "true" -a "$NUMERIC" != "false" && NUMERIC=$NUMERIC_DEFAULT
DENOM=1000000	# the denominator for the random fractional numbers

# FULL="true": full basis functions, FULL="false": interior basis functions
FULL_DEFAULT="true"
test "$FULL" != "true" -a "$FULL" != "false" && FULL=$FULL_DEFAULT

# SAGE="true": use SageMath, SAGE="false": use Mathematica
# use SageMath by default if it's available
if type sage >/dev/null 2>&1; then
    SAGE_DEFAULT="true"
else
    SAGE_DEFAULT="false"
fi
test "$SAGE" != "true" -a "$SAGE" != "false" && SAGE=$SAGE_DEFAULT

#---------------- determine the pattern of the uniform grid -----------------

get_orbit () {
    local i order dim

    dim=$1
    order=$2
    shift 2

    for ((i=0; i<=order; i++)); do count[i]=0; done
    for i in "$@"; do count[i]=$((count[i]+1)); done
    for ((i=$#; i<=dim; i++)); do count[0]=$((count[0]+1)); done
    pattern=`for ((i=0; i<=order; i++)); do echo ${count[i]}; done \
		| sort -nr | grep -v '^0$'`
    i=`echo $pattern | sed -e 's/ //g'`
    S[i]=$((S[i]+1))
}

uniform_orbits () {
    local i j k l m eq order dim

    dim=$1
    if $FULL; then eq="-ge"; else eq="-eq"; fi

    # Vertex node
    if test $dim $eq 0; then
	order=$2
	get_orbit $dim $order $order
    fi

    if test $dim $eq 1; then
	# Edge nodes
	order=$3
	for ((i=(order+1)/2; i<order; i++)); do
	    get_orbit $dim $order $i $((order-i))
	done
    fi

    if test $dim $eq 2; then
	# face nodes
	order=$4
	for ((i=(order+2)/3; i<order; i++)); do
	    for ((j=(order-i+1)/2; j<order-i && j<=i; j++)); do
		get_orbit $dim $order $i $j $((order-i-j))
	    done
	done
    fi

    if test $dim $eq 3; then
	# volume nodes
	order=$5
	for ((i=(order+3)/4; i<order; i++)); do
	    for ((j=(order-i+2)/3; j<order-i && j<=i; j++)); do
		for ((k=(order-i-j+1)/2; k<order-i-j && k<=j; k++)); do
		    get_orbit $dim $order $i $j $k $((order-i-j-k))
		done
	    done
	done
    fi

    if test $dim $eq 4; then
	# 4d nodes
	order=$6
	for ((i=(order+4)/5; i<order; i++)); do
	    for ((j=(order-i+3)/4; j<order-i && j<=i; j++)); do
		for ((k=(order-i-j+2)/3; k<order-i-j && k<=j; k++)); do
		    for ((l=(order-i-j-k+1)/2; l<order-i-j-k && l<=k; l++)); do
			get_orbit $dim $order $i $j $k $l $((order-i-j-k-l))
		    done
		done
	    done
	done
    fi
}

uniform_pattern () {
    local dim

    dim=$1
    if test $dim -eq "1"; then
	S[2]=0; S[11]=0
	uniform_orbits "$@"
	echo ${S[2]} ${S[11]}
    elif test $dim -eq "2"; then
	S[3]=0; S[21]=0; S[111]=0
	uniform_orbits "$@"
	echo ${S[3]} ${S[21]} ${S[111]}
    elif test $dim -eq "3"; then
	S[4]=0; S[31]=0; S[22]=0; S[211]=0; S[1111]=0
	uniform_orbits "$@"
	echo ${S[4]} ${S[31]} ${S[22]} ${S[211]} ${S[1111]}
    elif test $dim -eq "4"; then
	S[5]=0; S[41]=0; S[32]=0; S[311]=0; S[221]=0; S[2111]=0; S[11111]=0
	uniform_orbits "$@"
	echo ${S[5]} ${S[41]} ${S[32]} ${S[311]} ${S[221]} ${S[2111]} \
	     ${S[11111]}
    else
	echo 1>&2 "Invalid dimension: $dim"
	exit 1
    fi
}

#----------------------------------------------------------------------------

do_vertex () {
    if $first1; then first1="false"; else echo -n ","; fi
    echo -n "($1)^$order"
}

do_edge () {
    local i j
    for ((i=1; i<order; i++)); do
	if $first1; then first1="false"; else echo -n ","; fi
	j=$((order-i))
	echo -n "($1)^$i*($2)^$j"
    done
}

do_face () {
    local i j k
    for ((i=1; i<order-1; i++)); do
	for ((j=1; j<order-i; j++)); do
	    if $first1; then first1="false"; else echo -n ","; fi
	    k=$((order-i-j))
	    echo -n "($1)^$i*($2)^$j*($3)^$k"
	done
    done
}

do_volume () {
    local i j k l
    for ((i=1; i<order-2; i++)); do
	for ((j=1; j<order-i-1; j++)); do
	    for ((k=1; k<order-i-j; k++)); do
		if $first1; then first1="false"; else echo -n ","; fi
		l=$((order-i-j-k))
		echo -n "($1)^$i*($2)^$j*($3)^$k*($4)^$l"
	    done
	done
    done
}

do_4d () {
    local i j k l m
    for ((i=1; i<order-3; i++)); do
	for ((j=1; j<order-i-2; j++)); do
	    for ((k=1; k<order-i-j-1; k++)); do
		for ((l=1; l<order-i-j-k; l++)); do
		    if $first1; then first1="false"; else echo -n ","; fi
		    m=$((order-i-j-k-l))
		    echo -n "($1)^$i*($2)^$j*($3)^$k*($4)^$l*($5)^$m"
		done
	    done
	done
    done
}

#---------------------------- permutation macros -----------------------------

#---------------------------- 1D

perm_s11 () {		# perm_s11 cmd a b
    $1 "$2" "$3"
    $1 "$3" "$2"
}

#---------------------------- 2D

perm_s21 () {		# perm_s21 cmd a b
    perm_s11 "$1 $2" "$2" "$3"
    $1 "$3" "$2" "$2"
}

perm_s111 () {		# perm_s111 cmd a b c
    perm_s11 "$1 $2" "$3" "$4"
    perm_s11 "$1 $3" "$2" "$4"
    perm_s11 "$1 $4" "$3" "$2"
}

#---------------------------- 3D

perm_s31 () {		# perm_s31 cmd a b
    perm_s21 "$1 $2" "$2" "$3"
    $1 "$3" "$2" "$2" "$2"
}

perm_s22 () {		# perm_s22 cmd a b
    perm_s21 "$1 $2" "$3" "$2"
    perm_s21 "$1 $3" "$2" "$3"
}

perm_s211 () {		# perm_s211 cmd a b c
    perm_s111 "$1 $2" "$2" "$3" "$4"
    perm_s21 "$1 $3" "$2" "$4"
    perm_s21 "$1 $4" "$2" "$3"
}

perm_s1111 () {		# perm_s1111 cmd a b c d
    perm_s111 "$1 $2" "$3" "$4" "$5"
    perm_s111 "$1 $3" "$2" "$4" "$5"
    perm_s111 "$1 $4" "$2" "$3" "$5"
    perm_s111 "$1 $5" "$2" "$3" "$4"
}

#---------------------------- 4D

perm_s41 () {		# perm_s41 cmd a b
    perm_s31 "$1 $2" "$2" "$3"
    $1 "$3" "$2" "$2" "$2" "$2"
}

perm_s32 () {		# perm_s32 cmd a b
    perm_s22 "$1 $2" "$2" "$3"
    perm_s31 "$1 $3" "$2" "$3"
}

perm_s311 () {		# perm_s311 cmd a b c
    perm_s211 "$1 $2" "$2" "$3" "$4" 
    perm_s31 "$1 $3" "$2" "$4" 
    perm_s31 "$1 $4" "$2" "$3" 
}

perm_s221 () {		# perm_s221 cmd a b c
    perm_s211 "$1 $2" "$3" "$2" "$4"
    perm_s211 "$1 $3" "$2" "$3" "$4"
    perm_s22 "$1 $4" "$2" "$3"
}

perm_s2111 () {		# perm_s2111 cmd a b c d
    perm_s1111 "$1 $2" "$2" "$3" "$4" "$5"
    perm_s211 "$1 $3" "$2" "$4" "$5"
    perm_s211 "$1 $4" "$2" "$3" "$5"
    perm_s211 "$1 $5" "$2" "$3" "$4"
}

perm_s11111 () {	# perm_s11111 cmd a b c d e
    perm_s1111 "$1 $2" "$3" "$4" "$5" "$6"
    perm_s1111 "$1 $3" "$2" "$4" "$5" "$6"
    perm_s1111 "$1 $4" "$2" "$3" "$5" "$6"
    perm_s1111 "$1 $5" "$2" "$3" "$4" "$6"
    perm_s1111 "$1 $6" "$2" "$3" "$4" "$5"
}

basis2D () {
    # loops on basis functions for a given point ($1,$2,$3)

    if $first; then first="false"; else echo ","; fi

    first1="true"
    if $SAGE; then echo -n "["; else echo -n "{"; fi
    if ! $FULL; then
	# interior polynomial space \P_0(S_d;order)
	do_face "$1" "$2" "$3"
    else
	# full polynomial space \P(S_d;k0,k1,k2)
	order=$k0
	do_vertex "$1"		# vertex 0
	do_vertex "$2"		# vertex 1
	do_vertex "$3"		# vertex 2

	order=$k1
	do_edge "$1" "$2"	# edge 0-1
	do_edge "$1" "$3"	# edge 0-2
	do_edge "$2" "$3"	# edge 1-2

	order=$k2
	do_face "$1" "$2" "$3"	# face
    fi
    if $SAGE; then echo -n "]"; else echo -n "}"; fi
}

basis3D () {
    # loops on basis functions for a given point ($1,$2,$3,$4)

    if $first; then first="false"; else echo ","; fi

    first1="true"
    if $SAGE; then echo -n "["; else echo -n "{"; fi
    if ! $FULL; then
	# interior polynomial space \P_0(S_d;order)
	do_volume "$1" "$2" "$3" "$4"
    else
	# full polynomial space \P(S_d;k0,k1,k2,k3)
	order=$k0
	do_vertex "$1"		# vertex 0
	do_vertex "$2"		# vertex 1
	do_vertex "$3"		# vertex 2
	do_vertex "$4"		# vertex 4

	order=$k1
	do_edge "$1" "$2"	# edge 0-1
	do_edge "$1" "$3"	# edge 0-2
	do_edge "$1" "$4"	# edge 0-3
	do_edge "$2" "$3"	# edge 1-2
	do_edge "$2" "$4"	# edge 1-3
	do_edge "$3" "$4"	# edge 2-3

	order=$k2
	do_face "$1" "$2" "$3"	# face 3
	do_face "$1" "$2" "$4"	# face 2
	do_face "$1" "$3" "$4"	# face 1
	do_face "$2" "$3" "$4"	# face 0

	order=$k3
	do_volume "$1" "$2" "$3" "$4"	# volume
    fi
    if $SAGE; then echo -n "]"; else echo -n "}"; fi
}

basis4D () {
    # loops on basis functions for a given point ($1,$2,$3,$4,$5)

    if $first; then first="false"; else echo ","; fi

    first1="true"
    if $SAGE; then echo -n "["; else echo -n "{"; fi
    if ! $FULL; then
	# interior polynomial space \P_0(S_d;order)
	do_4d "$1" "$2" "$3" "$4" "$5"
    else
	# full polynomial space \P(S_d;k0,k1,k2,k3,k4)
	order=$k0
	do_vertex "$1"		# vertex 0
	do_vertex "$2"		# vertex 1
	do_vertex "$3"		# vertex 2
	do_vertex "$4"		# vertex 4
	do_vertex "$5"		# vertex 5

	order=$k1
	do_edge "$1" "$2"	# edge 0-1
	do_edge "$1" "$3"	# edge 0-2
	do_edge "$1" "$4"	# edge 0-3
	do_edge "$1" "$5"	# edge 0-4
	do_edge "$2" "$3"	# edge 1-2
	do_edge "$2" "$4"	# edge 1-3
	do_edge "$2" "$5"	# edge 1-4
	do_edge "$3" "$4"	# edge 2-3
	do_edge "$3" "$5"	# edge 2-4
	do_edge "$4" "$5"	# edge 3-4

	order=$k2
	do_face "$3" "$4" "$5"
	do_face "$2" "$4" "$5"
	do_face "$2" "$3" "$5"
	do_face "$2" "$3" "$4"
	do_face "$1" "$4" "$5"
	do_face "$1" "$3" "$5"
	do_face "$1" "$3" "$4"
	do_face "$1" "$2" "$5"
	do_face "$1" "$2" "$4"
	do_face "$1" "$2" "$3"

	order=$k3
	do_volume "$1" "$2" "$3" "$4"
	do_volume "$1" "$2" "$3" "$5"
	do_volume "$1" "$2" "$4" "$5"
	do_volume "$1" "$3" "$4" "$5"
	do_volume "$2" "$3" "$4" "$5"

	order=$k4
	do_4d "$1" "$2" "$3" "$4" "$5"
    fi
    if $SAGE; then echo -n "]"; else echo -n "}"; fi
}

filter () {
    # renames the random function
    if $SAGE; then
	sed -e 's/rand(\(.*\))\/'$DENOM'/random.randint(\1)\/'$DENOM';/g'
    else
	sed -e 's/rand(\(.*\))\/'$DENOM'/RandInt[{\1}]\/'$DENOM';/g'
    fi
}

unisolvence () {
    local i
    #
    # This function gengerates Mathematica commands for checking unisolvence of
    # a given orbit pattern with respect to \P_0(S_d;k). usage:
    #	unisolvence s0
    # or
    #	2D: unisolvence s3 s21 s111
    #	3D: unisolvence s4 s31 s22 s211 s1111
    #	4D: unisolvence s5 s41 s32 s311 s221 s2111 s11111
    # Note: dimension and order are stored in $dim and $order.
    #

    if $SAGE; then
	# defining variables for SageMath
        echo -n "ratpoly.<"
	nvar="0"
	if test "$dim" -eq "2" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i, b$i"
		nvar=$((nvar+2))
	    done
	elif test "$dim" -eq "2" -a "$#" -eq "3"; then
	    s3="$1"; s21="$2"; s111="$3"
	    for ((i=0; i<s21+s111; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s21; i<s21+s111; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "b$i"
		nvar=$((nvar+1))
	    done
	elif test "$dim" -eq "3" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i, b$i, c$i"
		nvar=$((nvar+3))
	    done
	elif test "$dim" -eq "3" -a "$#" -eq "5"; then
	    s4="$1"; s31="$2"; s22="$3"; s211="$4"; s1111="$5"
	    m=$((s31+s22+s211+s1111))
	    for ((i=0; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s31+s22; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "b$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s31+s22+s211; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "c$i"
		nvar=$((nvar+1))
	    done
	elif test "$dim" -eq "4" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i, b$i, c$i, d$i"
		nvar=$((nvar+4))
	    done
	elif test "$dim" -eq "4" -a "$#" -eq "7"; then
	    s5="$1"; s41="$2"; s32="$3"; s311="$4"; s221="$5"; s2111="$6"
	    s11111="$7"
	    m=$((s41+s32+s311+s221+s2111+s11111))
	    for ((i=0; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "a$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "b$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32+s311+s221; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "c$i"
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32+s311+s221+s2111; i<m; i++)); do
		test "$nvar" -gt "0" && echo -n ", "
		echo -n "d$i"
		nvar=$((nvar+1))
	    done
	else
	    echo 1>&2 "Invalid dimension (1)!"
	    exit 1
	fi
	test "$nvar" -eq "0" && echo -n "dummy"
        echo "> = QQ[]"
        ###echo "> = PolynomialRing(QQ,$nvar)"
        ###echo "> = PolynomialRing(GF(????))"
    fi

    if $NUMERIC; then
	# set random values to variables
	if $SAGE; then
	    echo "import random"
	else
	    echo "RandInt[{a_,b_}] := Round[a+Random[]*(b-a)];"
	fi
	nvar="0"
	if test "$dim" -eq "2" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" | filter
		nvar=$((nvar+2))
	    done
	elif test "$dim" -eq "2" -a "$#" -eq "3"; then
	    s3="$1"; s21="$2"; s111="$3"
	    for ((i=0; i<s21+s111; i++)); do
		if test $i -lt $s21; then
		    echo "a$i=rand(0,$((DENOM/2)))/$DENOM" | filter
		else
		    echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s21; i<s21+s111; i++)); do
		echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" | filter
		nvar=$((nvar+1))
	    done
	elif test "$dim" -eq "3" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" | filter
		echo "c$i=rand(0,$DENOM-(a$i+b$i)*$DENOM)/$DENOM" | filter
		nvar=$((nvar+3))
	    done
	elif test "$dim" -eq "3" -a "$#" -eq "5"; then
	    s4="$1"; s31="$2"; s22="$3"; s211="$4"; s1111="$5"
	    m=$((s31+s22+s211+s1111))
	    for ((i=0; i<m; i++)); do
	 	if test $i -lt $s31; then
		    echo "a$i=rand(0,$((DENOM/3)))/$DENOM" | filter
		elif test $i -lt $(( s31+s22+s211 )); then
		    echo "a$i=rand(0,$((DENOM/2)))/$DENOM" | filter
		else
		    echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s31+s22; i<m; i++)); do
	 	if test $i -lt $s211; then
		    echo "b$i=rand(0,$DENOM-2*a$i*$DENOM)/$DENOM" \
				| filter
		else
		    echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" \
				| filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s31+s22+s211; i<m; i++)); do
		echo "c$i=rand(0,$DENOM-(a$i+b$i)*$DENOM)/$DENOM" \
				| filter
		nvar=$((nvar+1))
	    done
	elif test "$dim" -eq "4" -a "$#" -eq "1"; then
	    s0="$1"
	    for ((i=0; i<s0; i++)); do
		echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" | filter
		echo "c$i=rand(0,$DENOM-(a$i+b$i)*$DENOM)/$DENOM" | filter
		echo "d$i=rand(0,$DENOM-(a$i+b$i+c$i)*$DENOM)/$DENOM" | filter
		nvar=$((nvar+4))
	    done
	elif test "$dim" -eq "4" -a "$#" -eq "7"; then
	    s5="$1"; s41="$2"; s32="$3"; s311="$4"; s221="$5"; s2111="$6"
	    s11111="$7"
	    m=$((s41+s32+s311+s221+s2111+s11111))
	    for ((i=0; i<m; i++)); do
	 	if test $i -lt $s41; then
		    echo "a$i=rand(0,$((DENOM/4)))/$DENOM" | filter
		elif test $i -lt $((s41+s32+s311)); then
		    echo "a$i=rand(0,$((DENOM/3)))/$DENOM" | filter
		elif test $i -lt $((s41+s32+s311+s221+s2111)); then
		    echo "a$i=rand(0,$((DENOM/2)))/$DENOM" | filter
		else
		    echo "a$i=rand(0,$DENOM)/$DENOM" | filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32; i<m; i++)); do
	 	if test $i -lt $((s41+s32+s311)); then
		    echo "b$i=rand(0,$DENOM-3*a$i*$DENOM)/$DENOM" | filter
		elif test $i -lt $((s41+s32+s311+s221)); then
		    echo "b$i=rand(0,$((DENOM/2))-a$i*$DENOM)/$DENOM" | filter
		elif test $i -lt $((s41+s32+s311+s221+s2111)); then
		    echo "b$i=rand(0,$DENOM-2*a$i*$DENOM)/$DENOM" | filter
		else
		    echo "b$i=rand(0,$DENOM-a$i*$DENOM)/$DENOM" | filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32+s311+s221; i<m; i++)); do
		if test $i -lt $((s41+s32+s311+s221+s2111)); then
		    echo "c$i=rand(0,$DENOM-(2*a$i+b$i)*$DENOM)/$DENOM" | filter
		else
		    echo "c$i=rand(0,$DENOM-(a$i+b$i)*$DENOM)/$DENOM" | filter
		fi
		nvar=$((nvar+1))
	    done
	    for ((i=s41+s32+s311+s221+s2111; i<m; i++)); do
		echo "d$i=rand(0,$DENOM-(a$i+b$i+c$i)*$DENOM)/$DENOM" | filter
		nvar=$((nvar+1))
	    done
	else
	    echo 1>&2 "Invalid dimension (2)!"
	    exit 1
	fi
    fi

    first="true"
    if $SAGE; then
	echo -n "A=matrix(["
    else
	echo -n "A={"
    fi

    if test "$#" -eq "1"; then
	s0="$1"
	if test "$dim" -eq "2"; then
	    for ((i=0; i<s0; i++)); do
		basis2D "a$i" "b$i" "1-a$i-b$i"
	    done
	elif test "$dim" -eq "3"; then
	    for ((i=0; i<s0; i++)); do
		basis3D "a$i" "b$i" "c$i" "1-a$i-b$i-c$i"
	    done
	elif test "$dim" -eq "4"; then
	    for ((i=0; i<s0; i++)); do
		basis3D "a$i" "b$i" "c$i" "d$i" "1-a$i-b$i-c$i-d$i"
	    done
	fi
    elif test "$dim" -eq "2" -a "$#" -eq "3"; then
	s3="$1"; s21="$2"; s111="$3"
	if $FULL; then
	    np=$(( 3+3*(k1-1)+(k2-1)*(k2-2)/2 ))
	else
	    np=$(( (order-1)*(order-2)/2 ))
	fi
	if test $(( s3+3*s21+6*s111 )) -ne $np; then
	    echo 1>&2 "order and orbits mismatch, abort."
	    exit 1
	fi

	# loop on the orbits
	test "$s3" -eq "1" && basis2D 1/3 1/3 1/3

	for ((i=0; i<s21; i++)); do
	    perm_s21 "basis2D" "a$i" "1-2*a$i"
	done

	for ((i=s21; i<s21+s111; i++)); do
	    perm_s111 "basis2D" "a$i" "b$i" "1-a$i-b$i"
	done
    elif test "$dim" -eq "3" -a "$#" -eq "5"; then
	s4="$1"; s31="$2"; s22="$3"; s211="$4"; s1111="$5"
	if $FULL; then
	    np=$(( 4+6*(k1-1)+4*(k2-1)*(k2-2)/2+(k3-1)*(k3-2)*(k3-3)/6 ))
	else
	    np=$(( (order-1)*(order-2)*(order-3)/6 ))
	fi
	if test $(( s4+4*s31+6*s22+12*s211+24*s1111 )) -ne $np; then
	    echo 1>&2 "order and orbits mismatch, abort."
	    exit 1
	fi

	# loop on the orbits
	test "$s4" -eq "1" && basis3D 1/4 1/4 1/4 1/4

	for ((i=0; i<s31; i++)); do
	    perm_s31 basis3D "a$i" "1-3*a$i"
	done

	for ((i=s31; i<s31+s22; i++)); do
	    perm_s22 basis3D "a$1" "1/2-a$i"
	done

	for ((i=s31+s22; i<s31+s22+s211; i++)); do
	    perm_s211 basis3D "a$i" "b$i" "1-2*a$i-b$i"
	done

	for ((i=s31+s22+s211; i<s31+s22+s211+s1111; i++)); do
	    perm_s1111 basis3D "a$i" "b$i" "c$i" "1-a$i-b$i-c$i"
	done
    elif test "$dim" -eq "4" -a "$#" -eq "7"; then
	s5="$1"; s41="$2"; s32="$3"; s311="$4"; s221="$5"; s2111="$6"
	s11111="$7"
	if $FULL; then
	    np=$(( 5+10*(k1-1)+10*(k2-1)*(k2-2)/2+5*(k3-1)*(k3-2)*(k3-3)/6
		+(k4-1)*(k4-2)*(k4-3)*(k4-4)/24 ))
	else
	    np=$(( (order-1)*(order-2)*(order-3)*(order-4)/24 ))
	fi
	if test $(( s5+5*s41+10*s32+20*s311+30*s221+60*s2111+120*s11111 )) \
		-ne $np; then
	    echo 1>&2 "order and orbits mismatch, abort."
	    exit 1
	fi

	# loop on the orbits

	# S5
	test "$s5" -eq "1" && basis4D 1/5 1/5 1/5 1/5 1/5

	# S41
	m=0
	for ((i=0; i<s41; i++)); do
	    perm_s41 basis4D "a$i" "1-3*a$i"
	done
	m=$((m+s41))

	# S32
	for ((i=m; i<m+s32; i++)); do
	    perm_s32 basis4D "a$i" "1/2-3*a$i/2"
	done
	m=$((m+s32))

	# S311
	for ((i=m; i<m+s311; i++)); do
	    perm_s311 basis4D "a$i" "b$i" "1-3*a$i-b$i"
	done
	m=$((m+s311))

	# S221
	for ((i=m; i<m+s221; i++)); do
	    perm_s221 basis4D "a$i" "b$i" "1-2*a$i-2*b$i"
	done
	m=$((m+s221))

	for ((i=m; i<m+s2111; i++)); do
	    perm_s2111 basis4D "a$i" "b$i" "c$i" "1-2*a$i-b$i-c$i"
	done
	m=$((m+s2111))

	for ((i=m; i<m+s11111; i++)); do
	    perm_s11111 basis4D "a$i" "b$i" "c$i" "d$i" "1-a$i-b$i-c$i-d$i"
	done
    else
	echo 1>&2 "Invalid dimension (3)!"
	exit 0
    fi

    if $SAGE; then
	echo "])"
	if $NUMERIC || test "$nvar" -eq "0"; then
	    opt=""
	else
	    opt="'frac'"	# faster
	fi
	###echo "print(simplify(det(A.echelon_form($opt))))"
	###echo "print(simplify(is_invertible(A.echelon_form($opt))))"
	cat <<END
res=1
for a in A.echelon_form($opt).diagonal():
    if a == 0:
	res = 0
print(res)
END
    else
	echo "};"
	echo "Simplify[Det[RowReduce[A]]]"
    fi
}

run_case () {
    local print 
    if $PRINT; then
	if $SAGE; then
	    echo 1>&2 "********* SageMath commands ***********"
	else
	    echo 1>&2 "********* Mathematica commands ***********"
	fi
	print="tee /dev/stderr"
    else
	print="cat"
    fi
    if $SAGE; then
	# evaluate using SageMath
	tmp="quadrule-ml-tmp-$$.sage"
	trap "/bin/rm -f $tmp $tmp.py" 0 1 2 3 15
	unisolvence "$@" | $print >$tmp
	res=`sage $tmp` #2>/dev/null`
	ret=$?
	/bin/rm -f $tmp $tmp.py
    else
	# evaluate using Mathematica
	res=`unisolvence "$@" | $print | math \
			| grep '^Out' | tail -1 | sed -e 's/ //g' \
			| awk -F= '{print $2}'`
	ret=$?
    fi
    if test -z "$res"; then
	echo "error!"
    elif test "$res" = "0"; then
        echo "[01;34mno[0m"
    else
	echo "[01;31myes[0m"
    fi
    return $ret
}

usage () {
    echo 1>&2 ""
    echo 1>&2 "Usage:"
    if $UNSYM; then		#----------------------------------------------
	if $FULL; then
	    echo 1>&2 "    % `basename $0` dim k0 ... k_{dim}"
	else
	    echo 1>&2 "    % `basename $0` dim order"
	fi
    elif $FULL; then		#----------------------------------------------
	cat 1>&2 <<END
    % `basename $0` dim orders [s3 s21 s111]

Examples:
    % `basename $0` 2 k0 k1 k2 [s3 s21 s111]
    % `basename $0` 3 k0 k1 k2 k3 [s4 s31 s22 s211 s1111]
    % `basename $0` 4 k0 k1 k2 k3 k4 [s5 s41 s32 s311 s221 s2111 s11111]
END
    else			#----------------------------------------------
	cat 1>&2 <<END
    % `basename $0` dim order [s3 s21 s111]

Examples:
    % `basename $0` 2 order [s3 s21 s111]
    % `basename $0` 3 order [s4 s31 s22 s211 s1111]
    % `basename $0` 4 order [s5 s41 s32 s311 s221 s2111 s11111]
END
    fi				#----------------------------------------------
    #==========================================================================
    cat 1>&2 <<END

Environment variables:
   ----------------------------------------------------------------------------
    Name	"true"			"false"			Default
   ----------------------------------------------------------------------------
    SAGE	use SageMath		use Mathematica		$SAGE_DEFAULT
    FULL	full polynomial space	interior only		$FULL_DEFAULT
    NUMERIC	fractional numbers	symbolic		$NUMERIC_DEFAULT
    UNSYM	unsymmetric point-set	symmetric point-set	$UNSYM_DEFAULT
    PRINT	true ==> print the Sage/Mathematica scripts	$PRINT_DEFAULT
   ----------------------------------------------------------------------------
END
}

##############################################################################

set -o noglob	# disable filename expansion to avoid problems with '*'

if $UNSYM; then
    echo "Checking unisolvence of unsymmetric point-sets."
else
    echo "Checking unisolvence of symmetric point-sets."
fi

dim=$1
test "$dim" != "2" -a "$dim" != "3" -a "$dim" != "4" && usage && exit 1

if $UNSYM; then
    if $FULL; then
	test $# -ne $((dim+2)) && usage && exit 1
    else
	test $# -ne 2 && usage && exit 1
    fi
elif $FULL; then
    if test $dim -eq 2; then
	test $# -ne 4 -a $# -ne 7 && usage && exit 1
    elif test $dim -eq 3; then
	test $# -ne 5 -a $# -ne 10 && usage && exit 1
    else
	test $# -ne 6 -a $# -ne 13 && usage && exit 1
    fi
else
    if test $dim -eq 2; then
	test $# -ne 2 -a $# -ne 5 && usage && exit 1
    elif test $dim -eq 3; then
	test $# -ne 2 -a $# -ne 7 && usage && exit 1
    else
	test $# -ne 2 -a $# -ne 9 && usage && exit 1
    fi
fi

if $FULL; then
    dim=$1; shift
    k0="$1"; shift
    k1="$1"; shift
    k2="$1"; shift
    test "$dim" -ge 3 && k3="$1" && shift
    test "$dim" -ge 4 && k4="$1" && shift
else
    dim=$1; shift
    order=$1; shift
fi

if test $# -ne 0; then
    run_case "$@"
    exit $?
fi 

if $SAGE; then
    echo "  . using SageMath for symbolic computations"
else
    echo "  . using Mathematica for symbolic computations"
fi

if $NUMERIC; then
    echo "  . check with randomly generated fractional numbers"
else
    echo "  . check with exact symbolic evaluations"
fi

if test $dim -eq 2; then
    if $FULL; then
	n=$(( 3+3*(k1-1)+(k2-1)*(k2-2)/2 ))
    else
	n=$(( (order-1)*(order-2)/2 ))
    fi
elif test $dim -eq 3; then
    if $FULL; then
	n=$(( 4+6*(k1-1)+4*(k2-1)*(k2-2)/2+(k3-1)*(k3-2)*(k3-3)/6 ))
    else
	n=$(( (order-1)*(order-2)*(order-3)/6 ))
    fi
elif test $dim -eq 4; then
    if $FULL; then
	n=$(( 5+10*(k1-1)+10*(k2-1)*(k2-2)/2+5*(k3-1)*(k3-2)*(k3-3)/6
		+(k4-1)*(k4-2)*(k4-3)*(k4-4)/24 ))
    else
	n=$(( (order-1)*(order-2)*(order-3)*(order-4)/24 ))
    fi
fi

if $UNSYM; then
    if $FULL; then
	case $dim in
	    2) temp="orders: $k0 $k1 $k2" ;;
	    3) temp="orders: $k0 $k1 $k2 $k3" ;;
	    4) temp="orders: $k0 $k1 $k2 $k3 $k4" ;;
	esac
    else
	temp="order: $order"
    fi
    echo "  . dim: $dim, $temp, number of points: $n"
    echo "Unisolvence: `run_case $n`"
elif test "$dim" -eq "2"; then
    echo "  . orbit patterns specified as:" S{3,21,111}
    if $FULL; then
	echo "  . dim: $dim, orders: $k0 $k1 $k2, number of points: $n"
	echo "  . orbit pattern of the uniform grid:" \
			"`uniform_pattern $dim $k0 $k1 $k2`"
    else
	echo "  . dim: $dim, order: $order, number of interior points: $n"
	echo "  . orbit pattern of the uniform grid:" \
			"`uniform_pattern $dim $order $order $order`"
    fi
    # n := s3 + 3*s21 + 6*s111
    s3=$(( n-(n/3)*3 ))
    n=$(( (n-s3)/3 ))
    # n := s21 + 2*s111
    for ((s21=n-(n/2)*2; s21<=n; s21+=2)); do
	s111=$(( (n-s21)/2 ))
	echo "	orbit pattern: $s3 $s21 $s111" \
			" - `run_case $s3 $s21 $s111`"
    done
elif test "$dim" -eq "3"; then
    echo "  . orbit patterns specified as:" S{4,31,22,211,1111}
    if $FULL; then
	echo "  . dim: $dim, orders: $k0 $k1 $k2 $k3, number of points: $n"
	echo "  . orbit pattern of the uniform grid:" \
			"`uniform_pattern $dim $k0 $k1 $k2 $k3`"
    else
	echo "  . dim: $dim, order: $order, number of interior points: $n"
	echo "  . orbit pattern of the uniform grid:" \
			"`uniform_pattern $dim $order $order $order $order`"
    fi
    # n := s4 + 4*s31 + 6*s22 + 12*s211 + 24*s1111
    s4=$(( n-(n/2)*2 ))
    n=$(( (n-s4)/2 ))
    # n := 2*s31 + 3*s22 + 6*s211 + 12*s1111
    for ((s22=n-(n/2)*2; 3*s22<=n; s22+=2)); do
	m=$(( (n-3*s22)/2 ))
	# m := s31 + 3*s211 + 6*s1111
	for ((s31=m-(m/3)*3; s31<=m; s31+=3)); do
	    l=$(( (m-s31)/3 ))
	    # l := s211 + 2*s1111
	    for ((s211=l-(l/2)*2; s211<=l; s211+=2)); do
		s1111=$(( (l-s211)/2 ))
		echo "	orbit pattern: $s4 $s31 $s22 $s211 $s1111" \
				" - `run_case $s4 $s31 $s22 $s211 $s1111`"
	    done
	done
    done
elif test "$dim" -eq "4"; then
    echo "  . orbit patterns specified as:" S{5,41,32,311,221,2111,11111}
    if $FULL; then
	echo "  . dim: $dim, orders: $k0 $k1 $k2 $k3 $k4," \
		"number of points: $n"
	echo "  . orbit pattern of the uniform grid:" \
			"`uniform_pattern $dim $k0 $k1 $k2 $k3 $k4`"
    else
	echo "  . dim: $dim, order: $order, number of interior points: $n"
	echo "  . orbit pattern of the uniform grid:" \
		"`uniform_pattern $dim $order $order $order $order $order`"
    fi
    # solve: s5+5*s41+10*s32+20*s311+30*s221+60*s2111+120*s11111 == n
    s5=$(( n%5 ))
    n=$(( n/5 ))
    # solve: s41+2*s32+4*s311+6*s221+12*s2111+24*s11111 == n
    for ((s41=n%2; s41<=n; s41+=2)); do
	m=$(( (n-s41)/2 ))
	# solve: s32+2*s311+3*s221+6*s2111+12*s11111 == m
	for ((s32=0; s32<=m; s32++)); do
	    l=$(( m-s32 ))
	    # solve: 2*s311+3*s221+6*s2111+12*s11111 == l
	    for ((s221=l%2; 3*s221<=l; s221+=2)); do
		k=$(( (l-3*s221)/2 ))
		# solve: s311+3*s2111+6*s11111 == k
		for ((s311=k%3; s311<=k; s311+=3)); do
		    j=$(( (k-s311)/3 ))
		    # solve: s2111+2*s11111 == j
		    for ((s2111=j%2; s2111<=j; s2111+=2)); do
			s11111=$(( (j-s2111)/2 ))
			pattern="$s5 $s41 $s32 $s311 $s221 $s2111 $s11111"
			echo "	orbit pattern: $pattern" \
				" - `run_case $pattern`"
		    done
		done
	    done
	done
    done
else
    echo 1>&2 "invalid dimension: $dim"
    exit 1
fi
